# VT Library v0.2

The VT library provides support for different serial protocols using the SAADC and PWM peripherals present on the nRF5x chips. The intent behind the library is to offer a way to use various serial protocols in a non-CPU bound manner, and without requiring specific hardware.

The main features of the VT library are:
- A generic way to receive and transmit serial data at bitrates up to 66 000,
in single-ended or differential mode.
- Transmission/receival of data that takes advantage of peripherals to reduce CPU load.

The library can be used through an existing frontend such as the UART frontend which simulates the UART peripheral, or it is possible to implement one’s own frontend to support a custom protocol.

![Block diagram of the modules in the VT library](documentation/images/vt_blocks.png)

The picture above shows how the VT library fits in an application.

---

# Table of contents

- [Requirements and compatibility](#requirements-and-compatibility)
- [SDK Library Dependencies](#sdk-library-dependencies)
- [SDK Library Conflicts](#sdk-library-conflicts)
- [Root file structure](#root-file-structure)
- [Installation](#installation)
  - [Windows](#installation-windows)
  - [Other / Manual](#installation-other)
    - [Installing the library in the SDK](#installation-other-library)
    - [Setting up the example projects](#installation-other-example-projects)
- [Hardware Setup](#hardware-setup)
  - [Using the library for UART over USB](#hardware-setup-uart-over-usb)
- [Included examples](#included-examples)
  - [Building with Segger Embedded Studio](#included-examples-build-segger)
  - [Using the pre-compiled .hex files](#included-examples-hex-files)
  - [Differences from ble\_app\_uart](#included-examples-differences)
- [Usage](#usage)
  - [Library configuration](#usage-library-configuration)
    - [Enabling the library](#usage-library-configuration-enabling)
      - [Additional configuration options](#additional-configuration-options)
    - [Base code requirements](#base-code-requirements)
      - [App scheduler](#base-code-requirements-app-sched)
      - [Initialization parameters](#initialization-parameters)
  - [Example: Using the UART frontend through the app\_uart module](#example-app-uart)
  - [Example: Using the UART frontend directly](#example-frontend-directly)
  - [Using the backend directly](#usage-backend-directly)
    - [Backend usage](#backend-usage)
      - [Receiving data](#backend-usage-receiving)
      - [Transmitting data](#backend-usage-transmitting)
      - [Payload formatting](#payload-formatting)
- [Notes on the sample rate](#notes-sample-rate)
- [RX buffers and improving RX responsiveness](#rx-buffers-responsiveness)
  - [Responsiveness example](#responsiveness-example)
- [Implementation details](#implementation-details)
  - [Peripheral configuration](#peripheral-configuration)
  - [RX buffer implementation](#rx-buffer-implementation)
  - [SAADC functionality](#saadc-functionality)
  - [TX buffer implementation and constructing the TX buffer](#tx-buffer-implementation)
  - [PWM functionality](#pwm-functionality)
- [Resource usage](#resource-usage)
  - [Backend flash usage](#backend-flash-usage)
  - [UART frontend flash usage](#uart-frontend-flash-usage)
  - [RX resource usage](#rx-resource-usage)
    - [Receiver memory usage](#receiver-memory-usage)
    - [Receiver CPU usage](#receiver-cpu-usage)
  - [TX resource usage](#tx-resource-usage)
    - [Transmitter memory usage](#transmitter-memory-usage)
    - [Transmitter CPU usage](#transmitter-cpu-usage)
- [Limitations and Issues](#limitations-and-issues)
  - [General](#limitations-general)
  - [UART frontend](#limitations-uart-frontend)
  - [Backend](#limitations-backend)
- [Additional notes](#additional-notes)
  - [Analog input pin numbers](#analog-pin-numbers)
  - [Manchester coding](#notes-manchester-coding)
- [API Reference](#api-reference)
- [Change log](#change-log)


---

# <a name="requirements-and-compatibility" /> Requirements and compatibility

The library has been tested with the following SoCs:
- nRF52832
- nRF52840

It may work with other nRF52 SoCs, but this is not guaranteed.

The VT library has been tested with the following nRF5 SDK versions:
- nRF5 SDK 15.0.0
- nRF5 SDK 15.2.0

[Download nRF5 SDK 15.2.0 here.](https://www.nordicsemi.com/eng/nordic/Products/nRF5-SDK/nRF5-SDK-zip/59011)

The example projects included in this repository require Segger Embedded Studio v3.40 or higher.
Segger Embedded Studio may be used for free with Nordic Semiconductor products. Follow the instructions on installation to obtain a license.

[Download Segger Embedded Studio here](https://www.segger.com/products/development-tools/embedded-studio/)

---

# <a name="sdk-library-dependencies" /> SDK Library Dependencies

The VT library depends on the following nRF5 SDK modules:
- **nrfx\_saadc** – for operating the SAADC. Use the included .c files as the SDK files have a slight incompatibility with the VT library.
- **nrfx\_pwm** – for operating the PWM
- **app\_scheduler** – for scheduling sample processing

**IMPORTANT**:  
Since the current **nrfx\_saadc** driver is incompatible with the VT library, you must instead
use the included files in the `modified_sdk` directory. Select the file with the file name that matches
your version of the nRF SDK, i.e. use **nrfx\_saadc_15.2.c** for SDK 15.2 and
**nrfx\_saadc_15.0.c** for SDK 15.0.

---

# <a name="sdk-library-conflicts" /> SDK Library Conflicts

If the macro `NRF_VT_REPLACE_NRFX_DRIVER=1`, the UART frontend will replace the UART nrfx drivers, and may not coexist with:
- **nrfx\_uart**
- **nrfx\_uarte**

---

# <a name="hardware-dependencies" /> Hardware dependencies

The VT library makes use of the following harware resources:
- **SAADC** (exclusive access)
- **PWM** (one module, configurable)
- **PPI** (one channel, configurable)


# Known issues

## nrfx\_saadc driver
The current version of the **nrfx\_saadc** driver does not fully support continuous
mode for the SAADC and will try to restart the SAADC using the CPU.
This can lead to samples being dropped between buffer swaps.
To avoid this the VT library uses PPI to restart the SAADC and uses a modified **nrfx\_saadc.c** which does not restart the SAADC. The VT library will be moved
to the default implementation once it is fixed.

## Power Management and sleeping

Using power management and different power modes is known to cause issues with the SAADC and the library code.
Avoid using such techniques to improve stability until the issues have been resolved in the library.

---


# <a name="root-file-structure" /> Root file structure

```
 .
 ├── README.md                 # README
 ├── setup_windows.exe         # Windows installation program
 ├── nrf_vt                    # VT library files
 ├── modified_sdk              # Contains modified nrfx_saadc.c
 ├── ble_app_uart_sdk_15.2.0   # Example project for SDK 15.2.0
 └── ble_app_uart_sdk_15.0.0   # Example project for SDK 15.0.0
```

---

# <a name="installation" />Installation

To install the library, first download this repository using `git clone` or in a .zip format.
Open the root directory of the repository and follow the instructions for your system below.

## <a name="installation-windows" /> Windows

Run the setup\_windows.exe file.
This will open a command line window where you must first select which SDK version you are using (15.0.0 and 15.2.0 supported).
Then, you will be prompted to select the root folder of the nRF5 SDK.
Once the root SDK directory has been selected, the installer will perform two actions:
- Symlink the library directory `nrf_vt` at `<nRF5 SDK root path>/components/libraries/nrf_vt`.
- Generate project files for the Segger Embedded Studio example projects included in this repository. Only the project files for the chosen SDK version are generated.

The Segger Embedded Studio project files will be located in the directories ending with sdk_15.2.0 for SDK 15.2.0 or sdk_15.0.0 for SDK 15.0.0,
at `PCA10040/s132/ses/example_name.emProject` for the nRF52832 or `PCA10056/s140/ses/example_name.emProject` for the nRF52840.

## <a name="installation-other" /> Other / Manual

If you are running a non-Windows operating system or the installation program does not work for you, it is possible to do the installation steps manually.

First, locate the path to the root directory of the nRF5 SDK you are using, for instance `/home/nordic/nRF5_SDK_15.2.0_9412b96` or `C:\Users\nordic\nRF5_SDK_15.2.0_9412b96`.

### <a name="installation-other-library" /> Installing the library in the SDK
You may then either:

symlink the library into the SDK directory by running:
```
ln -s nrf_vt <nRF5 SDK root path>/components/libraries/nrf_vt
```
on Linux / Mac OS, or
```
mklink /D <nRF5 SDK root path>/components/libraries/nrf_vt nrf_vt
```
on Windows,

or you can simply copy the `nrf_vt` directory into `<nRF5 SDK root path>/components/libraries/`.

### <a name="installation-other-example-projects" /> Setting up the example projects

If you wish to use the example projects provided in this repository without using the installer program, follow the steps below.
- First locate the example project for the SDK version you are using,
e.g. `ble_app_uart_sdk_15.2.0` for SDK version 15.2.0 or `ble_app_uart_sdk_15.0.0` for SDK version 15.0.0.
- In the directory for that example project, locate the project file directory for the SoC you are using, e.g.
`PCA10040/s132/ses` for nRF52832 or `PCA10056/s140/ses` for nRF52840.
- Copy the file from the `template` directory into the project file directory.
- Open the file with a text editor and locate the `macros=` field.
- Replace the text `REPLACE_ME_NOW=0` with `SDK_ROOT=<nRF5 SDK root path>` where `<nRF5 SDK root path>` should be replaced with the path to your nRF5 SDK.
	- Note: if you are on Windows, the backslashes in the path `\` must be substituted with forward slashes `/`.
- Save the file and rename it so that it has the extension **".emProject"**, e.g. rename `ble_app_uart_vt_pca10040_s132.template` to `ble_app_uart_vt_pca10040_s132.emProject`.

Now, the example project can be opened by opening the .emProject file in Segger Embedded Studio.

---

# <a name="hardware-setup" /> Hardware Setup

## <a name="hardware-setup-uart-over-usb" /> Using the library for UART over USB

In order to use the UART frontend to receive data sent over USB, the RX pin used by the VT library has to be connected to the ATMEL MCU UART TX pin.
This can be achieved by connecting a small wire from P0.08 (which is connected to the ATMEL MCU) to the configured RX pin (for instance P0.03/AIN1).

**Important**: In addition to this, the nRF52 chip may have to be cut off from the conductor path running from P0.08 to the MCU UART, otherwise it could interfere with the received signal.
You may omit this step, but at the risk of more bit errors.
See the Hardware Specification for your development kit for information on where to cut the path.

The examples included in this repository have the RX pin configured as P0.03/AIN1, so for these you would connect a wire from P0.08 to P0.03.

The PWM can use the UART TX pin P0.06 directly without requiring any additional intervention.

![nRF52 with wire running from P0.03 to P0.06](documentation/images/nrf52_uart_wiring.jpg)

See the picture for an example configuration.

---

# <a name="included-examples" /> Included examples

The repository includes some example applications that demonstrates how to integrate the VT library with BLE.  
The Segger Embedded Studio project files can be generated by performing the setup steps as described in [Installation](#installation).
To open a project, locate the Segger Embedded Studio project file at `<example directory>/pca10040/s132/ses/<example name>.emProject` for the PCA10040
or `<example directory>/pca10056/s140/ses/<example name>.emProject` for the PCA10056.
There are example projects for nRF5 SDK version 15.2.0 and nRF5 SDK version 15.0.0, so make sure to select the project that corresponds to your SDK version.

The projects provided are based on the existing SDK example project `ble_app_uart`.  
You can find detailed more documentation [here for the SDK 15.2.0 version](http://infocenter.nordicsemi.com/index.jsp?topic=%2Fcom.nordic.infocenter.sdk%2Fdita%2Fsdk%2Fnrf5_sdk.html)
, and [here for the SDK 15.0.0 version](http://infocenter.nordicsemi.com/index.jsp?topic=%2Fcom.nordic.infocenter.sdk%2Fdita%2Fsdk%2Fnrf5_sdk.html).

The SDK 15.0.0 projects use SoftDevice s132 v.6.0.0 for the PCA10040 and SoftDevice s140 v.6.0.0 for the PCA10056.  
The SDK 15.2.0 projects use SoftDevice s140 v.6.1.0 for the PCA10040 and SoftDevice s140 v.6.1.0 for the PCA10056.

## <a name="included-examples-build-segger" /> Building with Segger Embedded Studio

To build a project, click the `Build > Build <example name>` menu option in Segger Embedded Studio. After building, the build can be flashed to your development
kit at `Target > Download <example name>`.

## <a name="included-examples-hex-files" /> Using the pre-compiled .hex files

You may find pre-compiled .hex files in the `hex` directory of each project. These files have been merged with the relevant SoftDevice, and can be flashed directly to the development
kit using `nrfjprog` included in the [nRF5x Command Line tools](https://www.nordicsemi.com/eng/Products/Bluetooth-low-energy/nRF52-DK).

## <a name="included-examples-differences" /> Differences from ble\_app\_uart

The modifications made to the base projects are as follows:

The paths used in the project file were changed from relative paths to paths using the `SDK_ROOT` macro
(this isn't any special macro, but is just the name given to a macro we set to the path of the nRF5 SDK).

Defined the macros `NRF_VT_ENABLED=1`, `NRF_VT_FRONTEND=0`, `NRF_VT_REPLACE_NRFX_DRIVER=1`,
`NPWM_ENABLED=1`, `PWM0_ENABLED=1` and `SAADC_ENABLED=1` in the *Preprocessor definitions* in the project
(can also be found at the attribute "c_preprocessor_definitions" in the .emProject file).

Removed **nrfx\_uart.c**/**nrfx\_uarte.c** from the project and included
**nrfx\_saadc_15.x.c**, **nrfx\_pwm.c**, **nrf\_vt\_uart\_frontend.c** and **nrf\_vt\_backend.c**
(the file **app\_scheduler.c** was already present, but will have to be included in other projects).

Included the `app_scheduler` library in **main.c** and defined new RX and TX pins:
```
#if NRF_VT_ENABLED
#include "app_scheduler.h"

#define APP_SCHED_QUEUE_SIZE            8

#define VT_RX_PIN_NUMBER                3              /**< RX pin number to use when using VT for UART (AIN1). */
#define VT_TX_PIN_NUMBER                TX_PIN_NUMBER  /**< TX pin number to use when using VT for UART */

#endif // NRF_VT_ENABLED
```

**Note**: you may need to change the `VT_RX_PIN_NUMBER` and `VT_TX_PIN_NUMBER` depending on your setup.

Changed the device name:
```
#define DEVICE_NAME                     "Nordic_VT_UART"  /**< Name of device. Will be included in the advertising data. */
```

Changed the baud rate and RX/TX pins that were used to initialize the **app\_uart** module,
since the default 115200 is too high for the VT library:
```
static void uart_init(void)
{
    uint32_t                     err_code;
    app_uart_comm_params_t const comm_params =
    {
#if NRF_VT_ENABLED
        .rx_pin_no    = VT_RX_PIN_NUMBER,  // This is the new RX pin to use
        .tx_pin_no    = VT_TX_PIN_NUMBER,  // This is the new TX pin to use
#else
        .rx_pin_no    = RX_PIN_NUMBER,
        .tx_pin_no    = TX_PIN_NUMBER,
#endif
        .rts_pin_no   = RTS_PIN_NUMBER,
        .cts_pin_no   = CTS_PIN_NUMBER,
        .flow_control = APP_UART_FLOW_CONTROL_DISABLED,
        .use_parity   = false,
#if defined (UART_PRESENT)
#if NRF_VT_ENABLED
        .baud_rate    = NRF_UART_BAUDRATE_19200  // This is the new baudrate
#else
        .baud_rate    = NRF_UART_BAUDRATE_115200
#endif
#else
#if NRF_VT_ENABLED
        .baud_rate    = NRF_UARTE_BAUDRATE_19200  // This is the new baudrate
#else
        .baud_rate    = NRF_UARTE_BAUDRATE_115200
#endif
#endif
    };

    APP_UART_FIFO_INIT(&comm_params,
                       UART_RX_BUF_SIZE,
                       UART_TX_BUF_SIZE,
                       uart_event_handle,
                       APP_IRQ_PRIORITY_LOWEST,
                       err_code);
    APP_ERROR_CHECK(err_code);
}
```

Added initialization of **app\_scheduler** and added code to execute scheduled tasks in main():
```
int main(void)
{
    bool erase_bonds;

    // Initialize.
    uart_init();
    log_init();
    timers_init();
    buttons_leds_init(&erase_bonds);
    power_management_init();
    ble_stack_init();
    gap_params_init();
    gatt_init();
    services_init();
    advertising_init();
    conn_params_init();

#if NRF_VT_ENABLED
    APP_SCHED_INIT(0, APP_SCHED_QUEUE_SIZE);  // Here we initialize the app_scheduler
#endif

    // Start execution.
    printf("\r\nUART started.\r\n");
    NRF_LOG_INFO("Debug logging for UART over RTT started.");
    advertising_start();

    // Enter main loop.
    for (;;)
    {
#if NRF_VT_ENABLED
        app_sched_execute();   // Execute scheduled tasks in the main loop
#endif

        idle_state_handle();
    }
}
```

---

# <a name="usage" /> Usage

The VT library consists of a backend and one of several frontends.
The backend allows for generic data transmission/receival, using the nrfx_saadc and nrfx_pwm drivers to operate the SAADC and PWM peripherals respectively.
The frontend is where the protocol-specific implementation is present. It is intended to provide a higher-level interface to the application.

The existing UART frontend `nrf_vt_uart_frontend.c` can be used either directly (with an identical interface to the nrfx_uart/nrfx_uarte drivers) or it can be used through the `app_uart` library.

See [Implementation Details](#implementation-details) for more information.

## <a name="usage-library-configuration" /> Library configuration

### <a name="usage-library-configuration-enabling" /> Enabling the library

The following macros must be present to enable and use the library.
They may be placed in the `sdk_config.h` file or in the *Preprocessor Definitions* configuration section in Segger Embedded Studio:
- `NRF_VT_ENABLED`=1 - This enables the module
- `NRF_VT_FRONTEND`=0 - This enables the UART frontend (mandatory if using the UART frontend).
- `NRF_VT_REPLACE_NRFX_DRIVER`=1 - This makes the UART frontend use the same function names as the **nrfx\_uart**/**nrfx\_uarte** drivers (mandatory if using the UART frontend with app\_uart)

The **nrfx\_saadc** and **nrfx\_pwm** modules must also be enabled, which can be achieved by setting:
- `PWM_ENABLED`=1 - To enable **nrfx\_pwm**
- `PWM0_ENABLED`=1 - To enable usage of the PWM0 module
- `SAADC_ENABLED`=1 - To enable **nrfx\_saadc**

### <a name="additional-configuration-options"/>Additional configuration options

`NRF_VT_UART_PIN_N`  
This macro may be defined to make the UART frontend operate in differential mode.
It must be set to the pin number of an analog input, for instance *4* to use AIN2.
Once set, it is used as the negative pin in the SAADC differential mode.
The pin used for library initialization will then be regarded as the positive pin.

`NRF_VT_BACKEND_SAADC_CHANNEL`  
The SAADC channel number to use. Default: 0

`NRF_VT_BACKEND_RX_BUFFER_SIZE`  
The size (in samples) of each RX buffer. Default: 1024  
The total buffer space used will be equal to
`NRF_VT_BACKEND_RX_BUFFER_SIZE` * (2 + `NRF_VT_BACKEND_COPY_BUFFER_COUNT`) * sizeof(int16_t) Bytes

`NRF_VT_BACKEND_COPY_BUFFER_COUNT`  
The amount of reserve buffers to use for processing. Default: 2, Min: 1  
Increasing the amount of extra buffers may lead to less overrun errors due to buffer exhaustion, as it increases the minimum time interval during which processing must finish.

`NRF_VT_BACKEND_ADC_START_PPI_CHANNEL`  
The PPI channel used to trigger a TASK_START after EVENT_END on the SAADC. Default: 10  
This channel is required when a buffer is full to continue sampling into the next buffer.
May be set to any available PPI channel.

`NRF_VT_BACKEND_PWM_MODULE`  
The PWM module to use for transmission. Default: 0

`NRF_VT_BACKEND_PWM_BASE_CLOCK`  
The base clock rate used for the PWM. Default: NRF_PWM_CLK_8MHz  
May be set to any of the `nrf_pwm_clk_t` found in [`nrf_pwm.h`](http://infocenter.nordicsemi.com/index.jsp?topic=%2Fcom.nordic.infocenter.sdk%2Fdita%2Fsdk%2Fnrf5_sdk.html).
The COUNTERTOP value used for the PWM is set to `base clock rate / bitrate`.
You may need to change this value if using very high or very low bitrates.

`NRF_VT_BACKEND_TX_BUFFER_SIZE`  
The size (in PWM values) of the TX buffer. Default: 4096  
Increasing the TX buffer size will make it possible to send more data at once.
The total buffer space used will be equal to
`NRF_VT_BACKEND_TX_BUFFER_SIZE` * sizeof(uint16_t) Bytes

`NRF_VT_BACKEND_FRAME_CAST_COUNT`  
The amount of frame casts available. Default: 2

---

**The below options are only applicable if using the backend directly without an existing frontend**:

`NRF_VT_START_BIT_VALUE`  
The value of the start bit(s). Possible values: 0 / 1

`NRF_VT_START_BITS`  
The amount of start bits. Possible values: 1-8 (inclusive)

`NRF_VT_STOP_BIT_VALUE`  
The value of the stop bit(s). Possible values: 0 / 1

`NRF_VT_STOP_BITS`  
The amount of stop bits. Possible values: 1-8 (inclusive)

`NRF_VT_END_DELAY`  
Delay in bit times between each sent frame. Possible values: 1-8 (inclusive)

`NRF_VT_DEFAULT_LINE_VALUE`  
Idle line value. Possible values: 0 / 1

`NRF_VT_LSBIT_FIRST`  
Constant indicating whether the payload should be sent lsbit-first or msbit-first. Possible values: true / false


## <a name="base-code-requirements" /> Base code requirements

### <a name="base-code-requirements-app-sched" /> App scheduler
The VT library makes use of **app\_scheduler** to schedule processing of SAADC samples. To use this module:  
Include the header file by adding `#include "app_scheduler.h"`  
Initialize the scheduler and execute scheduled tasks in the main loop:  
```
int main(void)
{
	// ... other initialization
	APP_SCHED_INIT(/* Event Size */ 0, /* Queue size */ 8);

	// Main loop
	for (;;)
	{
		app_sched_execute(); // Add this line to your main loop
		//... other code
	}
}
```

The following snippet from the SAADC event handler in nrf_vt_backend.c shows the code which schedules the event:  
```
app_sched_event_put(NULL, 0, app_sched_process_handler);
```

This is done in an interrupt context. As can be seen, the event data is not used by the library.

The `app_sched_process_handler` is run once it reaches the front of the app scheduler queue.
Here is a snippet showing what it does:

```
/**@brief Event handler function for processing ADC samples */
static void app_sched_process_handler(void * p_event_data, uint16_t event_size)
{
    UNUSED_PARAMETER(p_event_data);
    UNUSED_PARAMETER(event_size);

    sample_buffer_node_t * p_node;

    while ((p_node = read_node_get()) != NULL)
    {
        samples_process(p_node->p_buffer, p_node->invalidate_after);
        read_pointer_move();
    }
}
```

As can be seen, once the processing is scheduled, the code will attempt to process any unprocessed buffers by moving the read pointer through the queue.

The event size and queue size used for the app scheduler can be different from in the example above,
however the queue size should be greater than the amount of RX copy buffers configured for the VT backend (default: 2).

### <a name="initialization-parameters" /> Initialization parameters

The following requirements must be met for the parameters used to initialize the module:
- The RX pin must be one of the Analog Inputs, i.e. AIN0-AIN7
- A negative RX pin may be provided to use differential mode. If provided, it must also
be an Analog Input and must be different from the positive RX pin.
- If the RX is not disabled, the bitrate must be no greater than 66 000 bps.
- If the RX is disabled, the bitrate may be higher. If using a very high or very low bitrate for transmission, you may need to change `NRF_VT_BACKEND_PWM_BASE_CLOCK`.
- The TX pin may be any of the GPIO pins on the board.
- The IRQ priority passed to the UART frontend or the priorities passed to the backend may be any valid priority, though if using the SoftDevice it is recommended to use `APP_IRQ_PRIORITY_LOW`.


## <a name="example-app-uart" /> Example: Using the UART frontend through the **app\_uart** module

The UART frontend functions in the same way as the **nrfx\_uart**/**nrfx\_uarte** drivers and can therefore be used through the **app\_uart** library,
which provides a simpler interface and a built-in FIFO.
There are some requirements set the VT UART frontend:
- Make sure that the macro `NRF_VT_REPLACE_NRFX_DRIVER` is set to `1` and that the aforementioned drivers are not included in the build as that will cause naming conflicts.
- If the receiver pin is enabled, ensure that you use a baudrate between `NRF_UART_BAUDRATE_1200` and `NRF_UART_BAUDRATE_19200` (inclusive), as that is the range that can be handled by the SAADC.
- If the receiver pin is enabled, also ensure that it is set to one of the AIN0-AIN7 pins. See [here](#analog-pin-numbers) for a list of pin numbers.

Remember to include the **nrfx\_saadc**, **nrfx\_pwm** and **app\_scheduler** dependencies,
as well as **nrf\_vt\_uart\_frontend**, **nrf\_vt\_backend** and of course **app\_uart**.

The below snippet shows an example usage of the **app\_uart** library with the VT UART frontend:

```
// ... other includes

#include "app_uart.h"

#if defined (UART_PRESENT)
#include "nrf_uart.h"
#endif
#if defined (UARTE_PRESENT)
#include "nrf_uarte.h"
#endif

#include "app_scheduler.h"

#define VT_RX_PIN_NUMBER 	3 	// AIN1
#define VT_TX_PIN_NUMBER 	25	// Some arbitrary unused pin

#define UART_TX_BUF_SIZE    256 // UART TX buffer size
#define UART_RX_BUF_SIZE    256 // UART RX buffer size


//... other code


/** Function for handling app_uart events. */
void uart_event_handle(app_uart_evt_t * p_event)
{
    switch (p_event->evt_type)
    {
        case APP_UART_DATA_READY:
            // Handle incoming data
            break;

        case APP_UART_COMMUNICATION_ERROR:
            // Handle communication errors
            break;

        case APP_UART_FIFO_ERROR:
            // Handle FIFO errors
            break;

        default:
            break;
    }
}


/** Function for initializing the UART module. */
static void uart_init(void)
{
    uint32_t                     err_code;
    app_uart_comm_params_t const comm_params =
    {
        .rx_pin_no    = VT_RX_PIN_NUMBER,
        .tx_pin_no    = VT_TX_PIN_NUMBER,
        .rts_pin_no   = RTS_PIN_NUMBER,
        .cts_pin_no   = CTS_PIN_NUMBER,
        .flow_control = APP_UART_FLOW_CONTROL_DISABLED,
        .use_parity   = false,
#if defined(UART_PRESENT)
        .baud_rate    = NRF_UART_BAUDRATE_19200
#else
		.baud_rate	  = NRF_UARTE_BAUDRATE_19200
#endif
	};

    APP_UART_FIFO_INIT(&comm_params,
                       UART_RX_BUF_SIZE,
                       UART_TX_BUF_SIZE,
                       uart_event_handle,
                       APP_IRQ_PRIORITY_LOWEST,
                       err_code);
    APP_ERROR_CHECK(err_code);
}


/** Application main function. */
int main(void)
{
    // Initialize the 'UART'
    uart_init();

	// ... other initialization code

	// Important: initialize the app_scheduler
    APP_SCHED_INIT(/* Event size */ 0, /* Queue size */ 8);

    // Enter main loop.
    for (;;)
    {
		// Important: this must be run in main() to process incoming serial data
        app_sched_execute();

		// ... other code
    }
}
```

For a more complete example, see the included example projects. These demonstrate use of **app\_uart** and the NUS service.

---

## <a name="example-frontend-directly" /> Example: Using the UART frontend directly

The UART frontend can be used directly without the **app\_uart** library.
Again, there are some requirements set the VT UART frontend:
- If the receiver pin is enabled, ensure that you use a baudrate between `NRF_UART_BAUDRATE_1200` and `NRF_UART_BAUDRATE_19200` (inclusive), as that is the range that can be handled by the SAADC with the given [limitations](#limitations).
- If the receiver pin is enabled, also ensure that it is set to one of the AIN0-AIN7 pins. See [here](#analog-pin-numbers) for a list of pin numbers
- By default the frontend will use the provided receiver pin in single-ended mode, where
the negative pin is internally shorted to ground on the PCB. If you wish to use differential mode, define the `NRF_VT_UART_PIN_N` macro to the negative pin number you want to use.

You may choose to set the macro `NRF_VT_REPLACE_NRFX_DRIVER` to `1` to change the prefix of the frontend API functions from 'nrf_vt_uart' to 'nrfx_uart',
however this is an optional feature when using the frontend directly.

Remember to include the **nrfx\_saadc**, **nrfx\_pwm** and **app\_scheduler** dependencies,
as well as **nrf\_vt\_uart\_frontend** and **nrf\_vt\_backend**.

The below snippet demonstrates usage of the VT UART frontend directly, with the `NRF_VT_REPLACE_NRFX_DRIVER` set to `1`.
If you wish to use **nrfx\_uart** and **nrf\_vt** at the same time:
- Set `NRF_VT_REPLACE_NRFX_DRIVER` set to `0`
- Change `#include "nrfx_uart.h"` to `#include "nrf_vt_uart_frontend.h"`
- Use the prefix 'nrf_vt_uart' instead of 'nrfx_uart' when using the VT UART frontend API, both for functions and constants.

```
// ... other includes

#include "nrfx_uart.h"

#if defined (UART_PRESENT)
#include "nrf_uart.h"
#endif
#if defined (UARTE_PRESENT)
#include "nrf_uarte.h"
#endif

/** Pin used for VT RX */
#define VT_RX_PIN_NUMBER	3  // AIN1
/** Pin used for VT TX */
#define VT_TX_PIN_NUMBER 	25 // Arbitrary unused pin

/** Size used for the VT RX buffers */
#define RX_BUFFER_SIZE (1)
/** Size used for the VT TX buffer */
#define TX_BUFFER_SIZE (1)

/** VT RX buffer A */
static uint8_t m_rx_buffer_a[RX_BUFFER_SIZE];
/** VT RX buffer B */
static uint8_t m_rx_buffer_b[RX_BUFFER_SIZE];

/** VT TX buffer */
static uint8_t m_tx_buffer[TX_BUFFER_SIZE];


// ... other code


/** Function for handling events from the nrf_vt module */
static void uart_event_handler(nrfx_uart_event_t const * p_event, void * p_context)
{
    switch (p_event->type)
    {
        case NRFX_UART_EVT_RX_DONE:
        {
			      // Handle the received data

			      // Set up the buffer for receiving again
            nrf_vt_uart_rx(NULL, p_event->data.rxtx.p_data, p_event->data.rxtx.bytes);
            break;
        }

        case NRFX_UART_EVT_TX_DONE:
        {
			     // Handle transmission done events
           break;
        }

        case NRFX_UART_EVT_ERROR:
        {
			      // Handle error events
            break;
        }
    }
}


/** Function for initializing the nrf_vt module */
static void uart_init(void)
{
    nrfx_err_t nrfx_err;

    nrfx_uart_config_t config = (nrfx_uart_config_t) NRFX_UART_DEFAULT_CONFIG;
#if defined(UARTE_PRESENT)
	config.baudrate = NRF_UART_BAUDRATE_19200;
#else
	config.baudrate = NRF_UARTE_BAUDRATE_19200;
#endif
	config.interrupt_priority = APP_IRQ_PRIORITY_LOW;
    config.pselrxd = VT_RX_PIN_NUMBER;
    config.pseltxd = VT_TX_PIN_NUMBER;

	// This function can be called with a NULL instead of a UART peripheral pointer
    nrfx_err = nrfx_uart_init(NULL, &config, uart_event_handler);
    if (nrfx_err != NRF_SUCCESS)
    {
		// Handle error
		return;
    }

	// Set up the two RX buffers for receiving
    nrfx_uart_rx(NULL, &m_rx_buffer_a[0], RX_BUFFER_SIZE);
    nrfx_uart_rx(NULL, &m_rx_buffer_b[0], RX_BUFFER_SIZE);
}


int main(void)
{
	// ... other initialization

	// Initialize the UART frontend
	uart_init();

	// Important: intialize the app_scheduler
    APP_SCHED_INIT(/* Event size */ 0, /* Queue size */ 8);

    for (;;)
    {
		// ... other code

		// Important: this must be run in main() to process incoming serial data
		app_sched_execute();

	}
}
```


## <a name="usage-backend-directly" /> Using the backend directly

In order to work without one of the provided frontends, the backend requires some that you provide it with some macros and functions.

The backend currently makes the assumption that all serial frames to be sent are in the format  
`Start Bits` `Payload` `Stop bits`.  
Additionally, the backend inserts between 1 and 8 bit times of delay between sent frames.
To configure the format of the frames and the end delay transmitted by the backend, use the below macros.  
**If used without a frontend, the backend requires that all the following macros exist**:

`NRF_VT_START_BIT_VALUE`  
`NRF_VT_START_BITS`  
`NRF_VT_STOP_BIT_VALUE`  
`NRF_VT_STOP_BITS`  
`NRF_VT_END_DELAY`  
`NRF_VT_DEFAULT_LINE_VALUE`  
`NRF_VT_LSBIT_FIRST`  

See [Additional configuration options](#additional-configuration-options) or the API documentation for the values to use for these macros.
The macros can be set in `sdk_config.h` or in the SES *Preprocessor Definitions* section.


**The backend also requires some functions to exist**:

---

```
bool vt_frontend_parse_next_bit(uint_fast8_t bit);
```

This function is called once for each bit received once a start bit is encountered, including for the first start bit.
It should return `true` if there are more bits required for the current frame or `false` otherwise (e.g. if the parser has not yet seen a start bit).

The processing code will try to minimize the processing required for idle times by only calling this function either
at the start of processing a new SAADC buffer or once it has encountered a start bit.
Therefore, remember to return `false` from this function whenever you want to wait for the next start bit.

The function will be called from a normal CPU context.

See `nrf_vt_uart_frontend.c` for an example of such a function.

---

```
void vt_frontend_overrun_handler(void);
```

This function is called after the processing code notices a loss in data due to buffer exhaustion.
It should handle this error condition with the knowledge that any amount of data may have been lost.
For instance, it may flush any partially constructed data.

The function will be called from a normal CPU context.

See `nrf_vt_uart_frontend.c` for an example of such a function.

---

```
void vt_frontend_tx_done_handler(void);
```

This function is called once the PWM has fully completed transmission of data.
The function should not perform any long-running tasks as it is called in an interrupt context.

The function will be called from an interrupt context at the interrupt priority set for the PWM in the backend initialization structure.

See `nrf_vt_uart_frontend.c` for an example of such a function.

### <a name="backend-usage" /> Backend usage

**Initialization**:  
The function `nrf_vt_backend_init` will initialize the backend.
It will enable the SAADC and PWM peripherals.
Example initialization using a baud rate of 19200 and single ended mode:
```
uint32_t err_code;

nrf_vt_backend_init_t config = {
	.baud_rate = 19200,									// The requested bit rate to receive / transmit at
	.rx_pin_p = 3,										// The positive RX pin (AIN1)
  .rx_pin_n = NRF_VT_PIN_DISCONNECTED, // The negative RX pin (disabled to use single-ended mode)
  .tx_pin = 25,										// The TX pin
	.pwm_interrupt_priority = APP_IRQ_PRIORITY_LOW,		// Interrupt priority used for PWM interrupts
	.saadc_interrupt_priority = APP_IRQ_PRIORITY_LOW,	// Interrupt priority used for SAADC interrupts
	.threshold_milli_volts = 1800						// Threshold value in millivolts. Used to differentiate low and high bits
};

err_code = nrf_vt_backend_init(&config);

// Check error code...
```
The `threshold_milli_volts` member is used to differentiate low and high bits - any sample under the threshold is regarded as a `0` and any sample above it is regarded as a `1`.
This mechanism is a bit too simplistic to use for some protocols, and will likely be changed in the near future.

The module can also be 'deinitialized' using `nrf_vt_uninit();` which will disable the peripherals and reset the backend to its initial state.


#### <a name="backend-usage-receiving" /> Receiving data

The following functions can be used to operate the receiver part of the backend:


Calling `nrf_vt_backend_rx_start();` will start the SAADC and begin receiving data and scheduling sample processing.

Calling `nrf_vt_backend_rx_stop();` will stop the SAADC so that sample processing is no longer scheduled.
Some processing tasks may still be left in the scheduler queue after calling this function.

`nrf_vt_backend_started()` lets you check if the receiver is running.

See the API Documentation for more information.

#### <a name="backend-usage-transmitting" /> Transmitting data  
Data transmission in the `nrf_vt` backend works similarly to many graphics systems.
First, write all the values to be transmitted to a transmission buffer, and then flush the transmission buffer by playing it back on the PWM.

**Simple writes**

One can call the function `nrf_vt_backend_tx_buffer_write_frame` to write to the transmission buffer. For example, to write a frame with a 16-bit payload to the buffer:  
```
uint8_t data[2] = {0x01, 0x02};

nrf_vt_backend_tx_buffer_write_frame(/* Data pointer */ &data[0], /* Number of data bits */ 16);
```

**Frame cast writes**

The `nrf_vt` library also includes a mechanism to perform more efficient transmission buffer construction.
It has so called "frame casts" which lets you create a buffer that is pre-filled with the start, stop and delay bits for a payload of a given length,
so that only the PWM values for the payload needs to be constructed. This reduces the copy function calls from 4 to 2.  

For example, preparing a frame cast for a payload of 16 bits:
```
nrf_vt_backend_frame_cast_init(/* Cast number */ 0, /* Number of payload bits */ 16);
```

Data can then be sent as such:
```
uint8_t data[2] = {0x01, 0x02};

nrf_vt_backend_tx_buffer_write_cast_frame(/* Cast number */ 0, &data);
```

The library keeps track of the payload bit count for each frame, so you don't need to supply it again.


**Other transmission API**

To clear data written to the transmission buffer, call
```
nf_vt_backend_tx_buffer_clear();
```

To start transmitting the transmission buffer, call  
```
nrf_vt_backend_tx_start();
```

After calling this function, you will not be able to write to the transmission buffer before the transmission is complete.

To stop an ongoing transmission, one can call
```
nrf_vt_backend_tx_stop();
```

Lastly, the function `nrf_vt_backend_tx_started` lets you check if a transmission is currently running.

See the API Documentation for more information.

#### <a name="payload-formatting" /> Payload formatting

**Least significant bit first**

For lsbit-first transmission, the data buffer supplied to the TX backend should be formatted as such:


```
data[N] = {
	Byte 0,
	Byte 1,
	...,
	Byte N-1
}
```

Each individual byte will be sent lsbit-first.

If the payload bit length is not a multiple of 8, place the `length % 8` bits in the least significant bits of the last byte in the data.

For example, we want to transmit a payload of length 20:

```
// Data we want to send (shown in msbit order): 0001 0010 0011 0100 0101
// Data sent on the serial line (lsbit-first order): 1010 0010 1100 0100 1000

uint8_t data[3] = {
	0b00010010,
	0b00110100,
	0b00000101  // The most significant 4 bits are ignored
};

nrf_vt_backend_tx_buffer_write_frame(/* Data pointer */ &data[0], /* Number of data bits */ 20);
```

**Most significant bit first**

For msbit-first transmission, the data supplied to the TX backend should be formatted as such:

```
data[N] = {
	Byte 0,
	Byte 1,
	...,
	Byte N-1
}
```

Each individual byte will be sent msbit-first.

If the payload bit length is not a multiple of 8, place the `length % 8` bits in the least significant bits of the last byte in the data.

For example, we want to transmit a payload of length 20:

```
// Data we want to send (shown in msbit-order): 0001 0010 0011 0100 0101
// Data sent on the serial line (msbit-first order): 0001 0010 0011 0100 0101

uint8_t data[3] = {
	0b00010010,
	0b00110100,
	0b00000101  // The most significant 4 bits are ignored
};

nrf_vt_backend_tx_buffer_write_frame(/* Data pointer */ &data[0], /* Number of data bits */ 20);
```

---

# <a name="notes-sample-rate" /> Notes on the sample rate

The `nrf_vt_backend` implementation uses between 3 and 7 samples per bit, meaning that the SAADC runs at
a higher sample rate than the bitrate.
The SAADC can sample at between 7816 Hz and 200 000 Hz, so the following equation must hold:  
`7 816 < bitrate * samples per bit <= 200 000 Hz`.

Based on a requested bitrate, the samples per bit will automatically be set to the lowest odd number between 3 and 7 that satisfies this equation.  
For instance, with a requested bitrate of *1200* bps, the samples per bit will be set to *7*.  
Or, with a requested bitrate of *19200* bps, the samples per bit will be set to *3*.

---

# <a name="rx-buffers-responsiveness" /> RX buffers and improving RX responsiveness

The receiver implementation will wait for an entire SAADC buffer to be filled before it is processed.
The time for one buffer to be filled is given by  
`T_buffer = Size_buffer * (1 / Samplerate)`

The sample buffer size can be changed using `NRF_VT_BACKEND_BUFFER_SIZE` (default: 1024).

The library uses multiple sample buffers organized in a queue. The buffers can be divided into the 'core' buffers
(reserved buffers required to keep the SAADC running continuously) and the 'copy' buffers (additional buffers). (See [here](#rx-buffer-implementation) for a detailed description of the buffers).
It is possible to configure the amount of buffers by changing `NRF_VT_BACKEND_COPY_BUFFER_COUNT` (default: 2).
Increasing the amount of copy buffers will increase the time window during which processing has to occur.

The time window during which processing a given buffer can be delayed is determined by  
`Max_delay = T_buffer * NRF_VT_BACKEND_COPY_BUFFER_COUNT`

If the library runs out of vacant copy buffers it will trigger an overrun error.

## <a name="responsiveness-example" /> Responsiveness example

Take for example a build running at 19 200 bps with the following configuration:
```
NRF_VT_BACKEND_BUFFER_SIZE=1024
NRF_VT_BACKEND_COPY_BUFFER_COUNT=2
```
That is, 19 200 bps with 3 samples/bit, with 2 copy buffers (= 4 buffers total with the 2 core buffers) each at 1024 samples size.
Based on this configuration, we have  
`Samplerate = 19 200 * 3 = 57 600`  
`T_buffer = 1024 * (1 / 57 600) = 0.0178s = 17.8ms`  
`Max_delay = 17.8ms * 2 = 35.6ms`

A typical cycle of sampling and processing will occur as shown in the graphic below.

![MSC showing two frames being sent](documentation/images/saadc_responsiveness_scenario1.png)

Here, `t_process` is the time it takes to process the samples, which will vary depending on how much data is being received.
From the timestamps in the graphic we can see that it takes over 12ms for frame 1
and over 15ms for frame 2 to be delivered to the Application Event Handler.
In many use cases this response time is too high for practical use.

As can be seen from the equation for `T_buffer` above we can reduce the response time by reducing the buffer size. However, by only reducing the buffer size
we decrease the maximum time window for processing `Max_delay`, increasing the risk of an overrun error. To compensate for this, we should increase the
amount of copy buffers.

We change the configuration to the following:
```
NRF_VT_BACKEND_BUFFER_SIZE=256
NRF_VT_BACKEND_COPY_BUFFER_COUNT=8
```

That is, 3 samples/bit, with 8 copy buffers (= 10 buffers total) each at 256 samples size.
Based on this configuration, we now have  
`T_buffer = 256 * (1 / 57 600) = 4.44ms`  
`Max_delay = 4.44ms * 8 = 35.5`

Since we increased the copy buffers by the same factor we decreased the buffer size, the `Max_delay` is basically unchanged. However, the `T_buffer` decreased by a factor of 4.
The new timing is shown in the graphic below.

![MSC showing two frames being sent](documentation/images/saadc_responsiveness_scenario2.png)

Now, we can see that frame 1 takes around 4ms and frame 2 takes about 3ms to be delivered to the Application Event Handler.
The tradeoff for changing this is that interrupts must be processed more often and
that the library therefore has a bigger CPU requirement and needs the CPU more often.

---

# <a name="implementation-details" />Implementation details

## <a name="peripheral-configuration" /> Peripheral configuration

The VT backend as of now runs with a set configuration for the SAADC and PWM peripherals:
Currently, the SAADC runs with the following configuration:
- Single-ended or differential mode depending on pin configuration
- 10 bit resolution
- No oversampling / burst
- Pulldown on the P-resistor
- Disabled N-resistor
- 1/6 input gain
- Internal reference voltage
- 10µs acquisition time
- Continuous sampling mode
- Internal timer set to the requested bitrate times the configured samples per bit: `CC = 16000000 / (bitrate * samples per bit)`

PPI is used to connect EVENT_END and TASK_START on the SAADC.


The PWM module in use runs with the following configuration:
- Up counter mode
- 8MHz base clock
- Common load mode
- Auto step mode
- Top value set based on the base clock to run at the requested bitrate: `Top_value = BASE_CLOCK / bitrate`


## <a name="rx-buffer-implementation" />RX buffer implementation

![Diagram showing a partially full RX buffer structure](documentation/images/vt_rx_buf_1.png)

The receiver has a special FIFO implementation designed to handle the need to always have a secondary buffer present in the SAADC RESULT double buffer, while at the same time preserving the FIFO property.
The implementation works like this:

- There is an array of sample buffers, `rx_buffers`. There are 2 + N buffers, where 2 are reserved for use as primary and secondary buffers for the SAADC and the remaining N are used for storing previously filled buffers that are ready for processing.
- There is an array of 2 + N 'nodes', `rx_nodes`. This array represents the FIFO queue used to keep track of the buffers to fill and process.
- Each node contains a pointer to a sample buffer and a flag `invalidate_after` (used to indicate when an overrun error has occurred).
- There are three pointers (indices) to the `rx_nodes` used for keeping track of the queue:
	- `rx_rd`, which points at the next node that should be processed
	- `rx_wr`, which points at the node (pointing to the buffer) currently in use as the primary buffer in the SAADC
	- `rx_wr_next`, which points at the node (pointing to the buffer) currently in use as the secondary buffer in the SAADC
- In normal conditions, the pointers move through the `rx_nodes` as in a circular buffer, `rx_wr_next` always being one step ahead of `rx_wr`.
	- The node pointed to by `rx_rd` can be processed as long as `rx_rd` != `rx_wr`.
	- Every time the primary buffer is filled, the `rx_wr` and `rx_wr_next` move one step ahead through the `rx_nodes` and the `rx_wr_next` buffer is set as the EasyDMA buffer in the SAADC.

When the processing lags behind, the queue can eventually be filled up. The graphic below shows that `rx_wr_next` has caught up with `rx_rd`, and an overrun is imminent.
This situation is resolved as follows:

![Diagram showing a full RX buffer structure, before the overrun handling](documentation/images/vt_rx_buf_2.png)

- The node immediately before the just finished primary buffer (as pointed to by `rx_wr`) is marked with the `invalidate_after` flag, indicating that there will be a loss of data after that node.
- The buffer pointers in the nodes pointed to by `rx_wr` and `rx_wr_next` are swapped.
- `rx_wr` and `rx_wr_next` stay unchanged.
- The `rx_wr_next` buffer is set as the EasyDMA buffer as usual

The effect of this is that the primary buffer previously used is recycled immediately as the secondary buffer,
while the chronological ordering of the buffers pointed to by the nodes in `rx_nodes` stays intact.
This way the reader can always process the next available node, making sure to invalidate its data after processing a buffer marked with `invalidate_after`.
The graphic below shows the buffer configuration after the overrun handling operation.

![Diagram showing a full RX buffer structure, after the overrun handing](documentation/images/vt_rx_buf_3.png)

## <a name="saadc-functionality" /> SAADC functionality

![MSC showing the basic operation cycle of the VT Receiver](documentation/images/saadc_usage.png)

The above sequence diagram shows an overview of how the SAADC is used and the processing takes place:
- Use the SAADC channel as defined by `NRF_VT_BACKEND_SAADC_CHANNEL` (Default: channel 0)
- On start, run a calibration task (`TASK_CALIBRATEOFFSET`)
- When calibration is done, set up the first `RESULT` buffer and start the ADC
- Then, start sampling and set up the second `RESULT` buffer
- When a sample buffer is filled:
  - The PPI automatically restarts the SAADC so that sampling is continuous. The SAADC will start using the previous secondary buffer (which has now become the primary buffer).
  - the SAADC event handler will set up the next `RESULT` buffer in the SAADC and schedule processing of the buffer just filled using **app\_scheduler**
- When scheduled, the processing handler will parse bits from any previously filled and unprocessed buffers
- The frontend in use will parse the incoming bits into frames of some specified format (in this case UART), and will deliver the data to the application once it has filled its buffers (which in the UART case are supplied by the application)
- Sampling and processing continue indefinitely or until an error condition occurs

## <a name="tx-buffer-implementation" /> TX buffer implementation and constructing the TX buffer

The transmitter has a single buffer used for storing the values fed to the PWM.
The size of this buffer can be set by changing `NRF_VT_BACKEND_TX_BUFFER_SIZE` (default: 4096).
Keep in mind that each bit will require one entry in the buffer, so the buffer size indicates the size in bits of the maximum transmission.
Also note that each frame will require start bits, stop bits and an end delay in the buffer, in addition to the payload bits themselves.

The transmitter has a lookup table which contains the PWM values representing each 8-bit value.
This lookup table is initialized based on whether the payload for the given protocol is lsbit-first or msbit-first.


## <a name="pwm-functionality" /> PWM functionality

![MSC showing the basic operation cycle of the VT transmitter](documentation/images/pwm_usage.png)

The above sequence diagram shows an overview of how the PWM is used.
- Use the PWM module as defined by `NRF_VT_BACKEND_PWM_MODULE` (default: 0)
- On initialization, a lookup buffer mapping 8-bit values to PWM sequence values is initialized
- Optionally, the backend user can predefine some 'frame casts' which are used to construct the PWM sequence values for the data frame with less copy operations than the default way
- The application calls some frontend-defined function with a data payload to initiate data transmission
- Each data frame is converted to PWM sequence values using the lookup buffer. Start and stop bits, along with an optional "end delay" between frames are predefined by the frontend in use. The payload bits are given by the aforementioned function call made by the application.  The PWM sequence values for each frame are written to a transmission buffer.
- Data is set up so that one PWM value is used for each bit in the data
- When every data frame has been written to the transmission buffer, transfer is started. The entire transmission buffer is played back by the PWM. During transmission, the frontend will not allow the application to initiate another transmission.

---

# <a name="Test results" /> Test results

The VT system has been tested with a serial loopback protocol, in which a package
consisting of 2-256 frames (0-254 of which are payload frames) is sent between the PC and nRF52 board (the nRF52 sends back an
  echo response to the echo request).

All tests below were made at 19 200 bps.  

#### Test 1

**Method**: Fixed package of `15 02 55 55 55 55 55 55 55 55 55 55 55 55 55 55 55 55 55 55 55 55`
**Runtime**: 14 hours

**Results**:  
**Total (31614 runs, 50400s): Sent 189684 packets, Got 189684 packets, Missing: 0 (Loss Rate = 0.0), Errors: 0 (Error Rate = 0.0)**

#### Test 2

The below test used randomized payloads between 0 and 30 frames long and ran for 14 hours.

**Method**: Package with randomized payload between 0 and 30 frames long.  
**Runtime**: 14 hours

**Results**:  
**Total (34018 runs, 50400s): Sent 204108 packets, Got 204108 packets, Missing: 0 (Loss Rate = 0.0), Errors: 0 (Error Rate = 0.0)**


#### Test 3

**Method**: Package with randomized payload with fixed length at 254 frames.  
**Runtime**: 15 minutes

**Results**:  
**Total (183 runs, 900s): Sent 1098 packets, Got 1098 packets, Missing: 0 (Loss Rate = 0.0), Errors: 0 (Error Rate = 0.0)**


# <a name="resource-usage" /> Resource usage

## <a name="backend-flash-usage" /> Backend flash usage
TBD


## <a name="uart-frontend-flash-usage" /> UART frontend flash usage
TBD


## <a name="rx-resource-usage" /> RX resource usage

### <a name="receiver-memory-usage" /> Receiver memory usage

The RX buffers in the receiver use up  
`Size_RX_Buffers = (2 + NRF_VT_BACKEND_COPY_BUFFER_COUNT) * NRF_VT_BACKEND_RX_BUFFER_SIZE * sizeof(int16_t)` Bytes.  
For example, a build with 2 copy buffers and buffer size 1024 would use up  
`(2 + 2) * 1024 * 2 = 8192` Bytes.

The other receiver variables have a negligible size.

### <a name="receiver-cpu-usage" /> Receiver CPU usage

The below data was gathered on a PCA10040,
using the UART frontend and the following configuration:

- Total sample rate: 57 600 (19200 bps, 3 samples/bit)
- RX buffer size: 128
- RX copy buffers: 8
- UART RX buffer size: 1
- Differential mode

The processing time tables shown below display processing time for the SAADC IRQ
and Sample Buffer processing with different amounts of bytes received in the sample buffer.
The processing time for a buffer depends on the amount of bytes received in
that buffer, because the frame parsing code is skipped for the idle bits.
Realistic values should lie somewhere between the low and high values.

Since the App Event Handler is called from the processing code, the CPU load
of that Handler will affect the time spent in processing. This test build uses
an empty handler that simply sets up the RX buffer for writing again.

The build used below is a debug build without optimizations and is intended to
provide conservative estimates.
A speedup will be had in a release build with optimizations.

#### Results:

**IRQ average runtime over 27325 cycles: 15.4us (0.70%)**

| Bytes / buffer | Process time for one RX buffer     | Cycles measured     |
| --------------------------------------------------------------------------|
| 0              | 52.1µs                             | 5715                |
| 1              | 85.9µs                             | 5382                |
| 2              | 123µs                              | 6287                |
| 3              | 162µs                              | 5448                |
| 4              | 189µs                              | 4493                |
|                |                                    |                     |
| **All**        | **120µs (5.4% CPU)**               | **27 325**          |


## <a name="tx-resource-usage" /> TX resource usage

### <a name="transmitter-memory-usage" /> Transmitter memory usage

The lookup table in the transmitter uses  
`Size_TX_Lookup = 256 * 8 * sizeof(uint16_t) = 256 * 8 * 2 = 4096` Bytes

The TX buffer itself uses  
`Size_TX_Buffer = NRF_VT_BACKEND_TX_BUFFER_SIZE * sizeof(uint16_t)`  
For example, a build with a TX buffer size of 4096 would use  
`4096 * 2 = 8192` Bytes

The other transmitter variables are negligible compared to these.


### <a name="transmitter-cpu-usage" /> Transmitter CPU usage

The below data was gathered on a PCA10040, using the following configuration:

- Bitrate: 19200 bps

The *frame cast* method was used for transmission (See [here](#backend-usage-transmitting)).
The build used below is a debug build without optimizations and is intended to
provide conservative estimates.
A speedup will be had in a release build with optimizations.

**Results**

| Time to write one frame (byte) to the TX buffer |
| ----------------------------------------------- |
| 8.5µs                                           |

---

# <a name="limitations-and-issues" /> Limitations and Issues

## <a name="limitations-general" /> General

- Due to limitations in the SAADC peripheral and the implementation of this library, the maximum bitrate is around `66 000`.
- The sample processing currently uses a single threshold value to determine whether a received measurement should be interpreted as a 0 or a 1. Changes will be introduced later to accommodate different protocols.
- Currently only one RX line and one TX line is supported. The library may be expanded later to allow for more than one RX/TX line.

## <a name="limitations-uart-frontend" /> UART frontend

The UART frontend attempts to duplicate the functionality of the **nrfx_uart** driver and the UART peripheral, however these have some functionality which is not yet present in the UART frontend.
This includes:
- 'Manual' (e.g. polling) mode
- Bit parity
- Hardware Flow Control (HWFC)
- Task/Event addresses (Tasks/Events are not available as there is no actual UART peripheral)

## <a name="limitations-backend" /> Backend

- The backend currently makes the assumption that all serial frames are in the format `Start Bits` `Payload` `Stop bits`
- The backend uses a single threshold voltage level to distinguish low and high values, and does not support LOW / HIGH "areas". This works for the UART case, but may be changed in the future.
- The PWM line will be low until the backend is initialized. This will cause a transition from low to high, which may be perceived as a start bit.

---

# <a name="additional-notes" /> Additional notes

## <a name="analog-pin-numbers" />Analog input pin numbers

The **PCA10040** has the following mapping between analog input numbers and pin numbers:

- AIN0: 2
- AIN1: 3
- AIN2: 4
- AIN3: 5
- AIN4: 28
- AIN5: 29
- AIN6: 30
- AIN7: 31

The PCA10056 should have roughly the same mapping, but please refer the Hardware Documentation
for that development kit.


## <a name="notes-manchester-coding" /> Manchester coding
Though Manchester coding is not implemented in the library, it is possible to write a frontend that handles this.
One should then run the receiver at twice the bitrate, to account for the half-bit encoding. For instance, if the signal is 1 200 bps, one should run at 2 400 bps to capture all the half-bit transitions.

---

# <a name="api-reference" /> API Reference

[The API reference can be found here](documentation/index.html)

---

# <a name="change-log" /> Change log

## <a name="change-v0.2.2" /> Version 0.2.2
- Minor cleanup
- After testing, determined that 3µs ACQ time can still be used with SAADC/PPI
	- Updated code/README to reflect this

## <a name="change-v0.2.1" /> Version 0.2.1

- Added support for differential mode
- Fixed the bit parsing algorithm
- Fixed an issue in the UART frontend preventing the secondary buffer from being utilized
- Fixed an issue with switching SAADC buffers which caused data loss between buffers

## <a name="change-v0.1" /> Version 0.1

Initial release
