/**
 * Copyright (c) 2015 - 2018, Nordic Semiconductor ASA
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 *    list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form, except as embedded into a Nordic
 *    Semiconductor ASA integrated circuit in a product or a software update for
 *    such product, must reproduce the above copyright notice, this list of
 *    conditions and the following disclaimer in the documentation and/or other
 *    materials provided with the distribution.
 *
 * 3. Neither the name of Nordic Semiconductor ASA nor the names of its
 *    contributors may be used to endorse or promote products derived from this
 *    software without specific prior written permission.
 *
 * 4. This software, with or without modification, must only be used with a
 *    Nordic Semiconductor ASA integrated circuit.
 *
 * 5. Any software provided in binary form under this license must not be reverse
 *    engineered, decompiled, modified and/or disassembled.
 *
 * THIS SOFTWARE IS PROVIDED BY NORDIC SEMICONDUCTOR ASA "AS IS" AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL NORDIC SEMICONDUCTOR ASA OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 * GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef NRF_VT_BACKEND_H__
#define NRF_VT_BACKEND_H__

#include <stdint.h>
#include <stdbool.h>

/**
 * @defgroup nrf_vt_backend VT backend
 * @{
 * @ingroup nrf_vt
 * @brief VT backend
 */

#define NRF_VT_PIN_DISCONNECTED (0) /**< Value used to indicate that an RX/TX pin should not be used */


/*
 *@brief Module initialization structure 
 */
typedef struct
{
    /**@brief Requested bitrate. The backend will automatically calculate the required samples/bit */
    uint32_t            baud_rate;
    /**@brief Positive RX pin */
    uint8_t             rx_pin_p;
    /**@brief Negative RX pin (may be set to NRF_VT_PIN_DISCONNECTED to use single-ended mode) */
    uint8_t             rx_pin_n;
    /**@brief TX pin */
    uint8_t             tx_pin;
    /**@brief Interrupt priority used for PWM interrupts */
    uint8_t             pwm_interrupt_priority;
    /**@brief Interrupt priority used for SAADC interrupts */
    uint8_t             saadc_interrupt_priority;
    /**@brief Threshold value in millivolts. Used to differentiate low and high bits */
    int16_t             threshold_milli_volts;
}
nrf_vt_backend_init_t;


/**
 * @brief Function for initializing the VT backend.
 *
 * This function configures and enables the SAADC and PWM using the provided configuration.
 *
 * @param[in] p_config      Pointer to the module initialization structure.
 *
 * @retval    NRF_SUCCESS				If the initialization was successful.
 * @retval    NRF_ERROR_INVALID_PARAM   If the module is passed invalid parameters in the configuration structure.
 * @retval    NRF_ERROR_INVALID_STATE   If the PWM or SAADC drivers detect a reinitialization.
 * @retval    NRF_ERROR_BUSY            If the PPI or SAADC channels are already in use.
 */
uint32_t nrf_vt_backend_init(nrf_vt_backend_init_t * p_config);


/**
 * @brief Function for uninitializing the VT backend.
 *
 * This function resets the configuration in the SAADC and PWM and the module itself.
 */
void nrf_vt_backend_uninit(void);


/**
 * @brief Function for starting the VT backend RX.
 *
 * This function starts (background) sampling on the SAADC.
 *
 * @retval	NRF_ERROR_SUCCESS	If the sampling was successfully started.
 * @retval	NRF_ERROR_BUSY		If the sampling is already running.
 */
uint32_t nrf_vt_backend_rx_start(void);


/**
 * @brief Function for stopping the VT backend RX.
 *
 * This function stops the background sampling on the SAADC.
 * The SAADC will return whatever samples were made before stopping,
 * so there may be some data left to process.
 */
void nrf_vt_backend_rx_stop(void);


/**
 * @brief Function for checking if the VT backend RX is running
 *
 * @retval true if the SAADC is sampling, false otherwise
 */
bool nrf_vt_backend_rx_started(void);


/**
 * @brief Function for writing a data frame to the TX buffer
 *
 * This function will write a series of values representing the data frame to the PWM sequence buffer.
 * It will run a bit slower than using @ref nrf_vt_backend_tx_buffer_write_cast_frame due to more copy operations.
 * To flush the sequence buffer onto the serial line, use @ref nrf_vt_backend_tx_start.
 *
 * @param[in]	p_data				Pointer to the payload data.
 * @param[in]	bits				Amount of bits in the payload data.
 *
 * @retval		NRF_SUCCESS			If the frame was successfully written to the PWM sequence buffer.
 * @retval		NRF_ERROR_NO_MEM	If there was not enough space in the PWM sequence buffer to write the frame.
 * @retval		NRF_ERROR_BUSY		If the PWM is currently transmitting the PWM sequence buffer.
 */
uint32_t nrf_vt_backend_tx_buffer_write_frame(const uint8_t const * p_data, uint32_t bits);


/**
 * @brief Function for initializing a frame cast.
 *
 * This function sets up a 'frame cast' for more efficient PWM sequence construction.
 * The cast will be pre-filled with start and stop bits so that only the payload bits need to be copied on construction.
 *
 * @param[in]	cast_no						The index of the cast to initialize.
 * @param[in]	payload_bits				The amount payload length in bits for the frame type this cast is used for.
 *
 * @retval		NRF_SUCCESS					If the initialization was successful.
 * @retval		NRF_ERROR_INVALID_PARAM  	If an invalid cast index or payload length was supplied.
 */
uint32_t nrf_vt_backend_tx_frame_cast_init(uint8_t cast_no, uint32_t payload_bits);


/**
 * @brief Function for writing a data frame to the TX buffer using a frame cast.
 *
 * This function will use the supplied frame cast to write a series of values representing the data
 * frame to the PWM sequence buffer. It will run faster than using @ref nrf_vt_backend_tx_buffer_write_frame.
 * To flush the sequence buffer onto the serial line, use @ref nrf_vt_backend_tx_start.
 *
 * @param[in]	cast_no				The index of the cast to use for frame construction.
 * @param[in]	p_data				Pointer to the payload data.
 *
 * @retval		NRF_SUCCESS			If the frame was successfully written to the PWM sequence buffer.
 * @retval		NRF_ERROR_NO_MEM	If there was not enough space in the PWM sequence buffer to write the frame.
 * @retval		NRF_ERROR_BUSY		If the PWM is currently transmitting the PWM sequence buffer.
 */
uint32_t nrf_vt_backend_tx_buffer_write_cast_frame(uint8_t cast_no, const uint8_t const * p_data);


/**
 * @brief Function for clearing the TX buffer.
 *
 * This function will empty the PWM sequence buffer.
 */
void nrf_vt_backend_tx_buffer_clear(void);


/**
 * @brief Function for starting transmission of the TX buffer
 *
 * This function will start flushing the PWM sequence buffer onto the serial line using PWM.
 * The function returns immediately and the PWM activity will run in the background until all
 * the data has been flushed.
 *
 * @retval		NRF_SUCCCESS		If the PWM sequence was started successfully.
 * @retval		NRF_ERROR_BUSY		If the PWM is already transmitting.
 */
uint32_t nrf_vt_backend_tx_start(void);


/**
 * @brief Function for checking if the VT backend TX is running
 *
 * @retval true if the PWM is playing a sequence, false otherwise
 */
bool nrf_vt_backend_tx_started(void);


/**
 * @brief Stop the VT backend TX
 */
void nrf_vt_backend_tx_stop(void);


/** @} */

#endif // NRF_VT_BACKEND_H__
